"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.computeScore = exports.WindowedMatcher = exports.TokensCache = void 0;
const types_1 = require("../types");
class TokensCache {
    constructor(size) {
        this.keys = [];
        this.cache = {};
        this.size = size;
    }
    put(id, tokens) {
        var _a;
        this.cache[id] = tokens;
        if (!this.keys.includes(id)) {
            this.keys.push(id);
        }
        if (this.keys.length > this.size) {
            const key = (_a = this.keys.shift()) !== null && _a !== void 0 ? _a : '';
            delete this.cache[key];
        }
    }
    get(id) {
        return this.cache[id];
    }
}
exports.TokensCache = TokensCache;
const englishStopWords = new Set([
    'we',
    'our',
    'you',
    'it',
    'its',
    'they',
    'them',
    'their',
    'this',
    'that',
    'these',
    'those',
    'is',
    'are',
    'was',
    'were',
    'be',
    'been',
    'being',
    'have',
    'has',
    'had',
    'having',
    'do',
    'does',
    'did',
    'doing',
    'can',
    'don',
    't',
    's',
    'will',
    'would',
    'should',
    'what',
    'which',
    'who',
    'when',
    'where',
    'why',
    'how',
    'a',
    'an',
    'the',
    'and',
    'or',
    'not',
    'no',
    'but',
    'because',
    'as',
    'until',
    'again',
    'further',
    'then',
    'once',
    'here',
    'there',
    'all',
    'any',
    'both',
    'each',
    'few',
    'more',
    'most',
    'other',
    'some',
    'such',
    'above',
    'below',
    'to',
    'during',
    'before',
    'after',
    'of',
    'at',
    'by',
    'about',
    'between',
    'into',
    'through',
    'from',
    'up',
    'down',
    'in',
    'out',
    'on',
    'off',
    'over',
    'under',
    'only',
    'own',
    'same',
    'so',
    'than',
    'too',
    'very',
    'just',
    'now',
]);
const programmingStopWords = new Set([
    'if',
    'then',
    'else',
    'for',
    'while',
    'with',
    'def',
    'function',
    'return',
    'TODO',
    'import',
    'try',
    'catch',
    'raise',
    'finally',
    'repeat',
    'switch',
    'case',
    'match',
    'assert',
    'continue',
    'break',
    'const',
    'class',
    'enum',
    'struct',
    'static',
    'new',
    'super',
    'this',
    'var',
]);
const allStopWords = new Set([...englishStopWords, ...programmingStopWords]);
const languageStopWords = new Map();
function splitIntoWords(text) {
    return text.split(/[^a-zA-Z0-9]/).filter((word) => word.length > 0);
}
class DocumentTokenizer {
    constructor(resource) {
        var _a;
        this.stopsForLanguage = (_a = languageStopWords.get(resource.languageId)) !== null && _a !== void 0 ? _a : allStopWords;
    }
    tokenize(text) {
        return new Set(splitIntoWords(text).filter((word) => !this.stopsForLanguage.has(word)));
    }
}
const tokensCache = new TokensCache(20);
class WindowedMatcher {
    constructor(doc) {
        this.referenceDoc = doc;
        this.tokenizer = new DocumentTokenizer(doc);
    }
    get referenceTokens() {
        if (!this.referenceTokensCache) {
            this.referenceTokensCache = new Set(this.tokenizer.tokenize(this.getCursorContextInfo(this.referenceDoc).context));
        }
        return this.referenceTokensCache;
    }
    sortScoredSnippets(snippets, sortRule = types_1.SortOption.Descending) {
        // eslint-disable-next-line no-nested-ternary
        return sortRule === types_1.SortOption.Ascending
            ? snippets.sort((snippet1, snippet2) => (snippet1.score > snippet2.score ? 1 : -1))
            : sortRule === types_1.SortOption.Descending
                ? snippets.sort((snippet1, snippet2) => (snippet1.score > snippet2.score ? -1 : 1))
                : snippets;
    }
    retrieveAllSnippets(resource, sortRule = types_1.SortOption.Descending) {
        var _a;
        const snippets = [];
        if (!resource.source.length || !this.referenceTokens.size) {
            return snippets;
        }
        const sourceArray = resource.source.split('\n');
        const key = `${this.id()}:${resource.source}`;
        const cache = (_a = tokensCache.get(key)) !== null && _a !== void 0 ? _a : [];
        const noCache = !cache.length;
        const tokens = noCache ? sourceArray.map((text) => this.tokenizer.tokenize(text), this.tokenizer) : [];
        for (const [index, [startLine, endLine]] of this.getWindowsDelineations(sourceArray).entries()) {
            if (noCache) {
                const size = new Set();
                tokens.slice(startLine, endLine).forEach((token) => token.forEach((word) => size.add(word)));
                cache.push(size);
            }
            const traget = cache[index];
            const score = this.similarityScore(traget, this.referenceTokens);
            snippets.push({
                score,
                startLine,
                endLine,
            });
        }
        if (noCache) {
            tokensCache.put(key, cache);
        }
        return this.sortScoredSnippets(snippets, sortRule);
    }
    findMatches(resource, mode = types_1.SnippetSelectionOption.BestMatch, size) {
        if (mode === types_1.SnippetSelectionOption.BestMatch) {
            const snippet = this.findBestMatch(resource);
            return snippet ? [snippet] : [];
        }
        return mode === types_1.SnippetSelectionOption.TopK ? this.findTopKMatches(resource, size) : [];
    }
    /**
     * 获取所有匹配的片段
     * @param resource 文档资源
     */
    findBestMatch(resource) {
        if (!resource.source.length || !this.referenceTokens.size) {
            return;
        }
        const lines = resource.source.split('\n');
        const snippets = this.retrieveAllSnippets(resource, types_1.SortOption.Descending);
        return snippets.length !== 0 && snippets[0].score !== 0
            ? Object.assign({ snippet: lines.slice(snippets[0].startLine, snippets[0].endLine).join('\n'), semantics: types_1.SnippetSemantics.Snippet }, snippets[0]) : null;
    }
    /**
     * 获取最高匹配片段
     * @param resource 文档资源
     * @param size 取前几位，默认取前一位
     */
    findTopKMatches(resource, size = 1) {
        if (!resource.source.length || !this.referenceTokens.size || size < 1) {
            return;
        }
        const lines = resource.source.split('\n');
        const snippets = this.retrieveAllSnippets(resource, types_1.SortOption.Descending);
        if (!snippets.length || !snippets[0].score) {
            return;
        }
        const topSnippets = [snippets[0]];
        for (let i = 1; i < snippets.length && topSnippets.length < size; i++) {
            if (topSnippets.findIndex((t) => snippets[i].startLine < t.endLine && snippets[i].endLine > t.startLine) === -1) {
                topSnippets.push(snippets[i]);
            }
        }
        return topSnippets.map((snippet) => (Object.assign({ snippet: lines.slice(snippet.startLine, snippet.endLine).join('\n'), semantics: types_1.SnippetSemantics.Snippet }, snippet)));
    }
}
exports.WindowedMatcher = WindowedMatcher;
function computeScore(set1, set2) {
    let intersectionSize = 0;
    for (const element of set1) {
        if (set2.has(element)) {
            intersectionSize++;
        }
    }
    return intersectionSize / (set1.size + set2.size - intersectionSize);
}
exports.computeScore = computeScore;
//# sourceMappingURL=matcher.js.map