"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getAfterPrompt = exports.getBeforePrompt = exports.getCroppedText = exports.getCroppedTextByLine = exports.getMarkerForSnippets = exports.getPathMarker = exports.getLanguageMarker = exports.getMarkerByLanguage = exports.addComment = void 0;
const service_1 = require("../../languages/service");
const types_1 = require("../types");
const const_1 = require("./const");
const tokenizer_1 = require("./tokenizer");
const addComment = (commentText) => {
    if (commentText) {
        return `${commentText}\n`;
    }
    return '';
};
exports.addComment = addComment;
const getMarkerByLanguage = (text, language) => {
    const marker = const_1.LANGUAGE_COMMENT_MARKERS[language];
    if (marker) {
        const end = marker.end === '' ? '' : ` ${marker.end}`;
        return `${marker.start} ${text}${end}`;
    }
    return '';
};
exports.getMarkerByLanguage = getMarkerByLanguage;
const getLanguageMarker = (language) => {
    if (!language) {
        return '';
    }
    const supportLanguage = {
        html: '<!DOCTYPE html>',
        python: '#!/usr/bin/env python3',
        ruby: '#!/usr/bin/env ruby',
        shellscript: '#!/bin/sh',
        yaml: '# YAML data',
    };
    let marker = '';
    if (supportLanguage[language]) {
        return supportLanguage[language];
    }
    marker = `Language: ${language}`;
    return (0, exports.getMarkerByLanguage)(marker, language);
};
exports.getLanguageMarker = getLanguageMarker;
// 根据语言类型获取不同的注释内容
const getPathMarker = (path, language) => path && (0, exports.getMarkerByLanguage)(`Path: ${path}`, language);
exports.getPathMarker = getPathMarker;
const getMarkerForSnippets = (text, language) => {
    const lines = text.split('\n');
    return lines.map((line) => (0, exports.getMarkerByLanguage)(line, language)).join('\n');
};
exports.getMarkerForSnippets = getMarkerForSnippets;
const getCroppedTextByLine = (text, maxTokenSize, textTokens, reverse = false) => {
    const currentTokenSize = textTokens.reduce((prev, cur) => prev + cur.length, 0);
    if (currentTokenSize < maxTokenSize) {
        return text;
    }
    const lines = text.split('\n');
    const endLine = lines.length;
    let currenSize = 0;
    if (reverse) {
        let index = 0;
        for (; index < endLine; index++) {
            const tokens = textTokens[index];
            currenSize += tokens.length;
            if (currenSize < maxTokenSize) {
                continue;
            }
            else {
                break;
            }
        }
        if (index === endLine) {
            return text;
        }
        return lines.slice(0, index).join('\n');
    }
    let index = endLine - 1;
    for (; index >= 0; index--) {
        const tokens = textTokens[index];
        currenSize += tokens.length;
        if (currenSize < maxTokenSize) {
            continue;
        }
        else {
            break;
        }
    }
    if (index === -1) {
        return text;
    }
    return lines.slice(index).join('\n');
};
exports.getCroppedTextByLine = getCroppedTextByLine;
/**
 * 裁剪字符函数
 * @param text 文本内容
 * @param maxTokenSize 最大 token 数量
 * @param strategy 分割策略
 * @param language 语言类型，按函数分割仅在 typescript/javascript/typescriptreact 中支持
 * @param reverse 是否反向裁剪，即从后往前裁剪
 */
const getCroppedText = async (text, maxTokenSize, textTokens, strategy = types_1.StrategyType.InterceptBasedOnLine, tokenizer, parser, minBlockSize = 20, reverse = false, token) => {
    let tokens;
    if (strategy === types_1.StrategyType.InterceptBasedOnLine) {
        // 按行进行裁剪
        text = (0, exports.getCroppedTextByLine)(text, maxTokenSize, textTokens, reverse);
    }
    else if (strategy === types_1.StrategyType.InterceptBasedOnFunction && !reverse) {
        tokens = tokenizer.encode(text);
        // 按函数进行裁剪
        while (tokens.length > maxTokenSize) {
            if (reverse) {
                try {
                    text = text.slice(0, Math.ceil((maxTokenSize / tokens.length) * tokens.length));
                    text = (await (parser === null || parser === void 0 ? void 0 : parser.trimSuffixSyntaxErrors(text, minBlockSize))) || text;
                }
                catch (_a) {
                    text = (0, exports.getCroppedTextByLine)(text, maxTokenSize, textTokens, reverse);
                }
            }
            else {
                try {
                    text = text.slice(-Math.ceil((maxTokenSize / tokens.length) * tokens.length));
                    text = (await (parser === null || parser === void 0 ? void 0 : parser.trimPrefixSyntaxErrors(text, minBlockSize))) || text;
                    if (token === null || token === void 0 ? void 0 : token.isCancellationRequested) {
                        return '';
                    }
                }
                catch (_b) {
                    text = (0, exports.getCroppedTextByLine)(text, maxTokenSize, textTokens, reverse);
                }
            }
            tokens = tokenizer.encode(text);
        }
    }
    else {
        tokens = tokenizer.encode(text);
        // 按字符进行裁剪
        while (tokens.length > maxTokenSize) {
            const splitWords = text.split('');
            if (reverse) {
                text = splitWords.slice(0, -(tokens.length - maxTokenSize)).join('');
            }
            else {
                text = splitWords.slice(tokens.length - maxTokenSize).join('');
            }
            tokens = tokenizer.encode(text);
        }
    }
    return text;
};
exports.getCroppedText = getCroppedText;
const getBeforePrompt = async (promptPriority, context, promptConfig, leftTokenSize, injector, token) => {
    var _a;
    const { tokenizerName, wishList } = promptConfig;
    const tokenizer = (0, tokenizer_1.getTokenizer)(tokenizerName);
    // 根据优先级降序排序
    const sortedPromptPriority = promptPriority.sort((a, b) => b.priority - a.priority);
    const promptList = {};
    for (const current of sortedPromptPriority) {
        if (!leftTokenSize) {
            return;
        }
        if (current.content) {
            const tokens = current.content.split('\n').map((line) => tokenizer.encode(line));
            const currentTokenSize = tokens.reduce((prev, cur) => prev + cur.length, 0);
            const maxTokenSize = Math.ceil(current.maxPercent * leftTokenSize);
            if (currentTokenSize > leftTokenSize) {
                const languageParserService = injector.get(service_1.LanguageParserService);
                const languageParser = languageParserService.createParser(context.language);
                promptList[current.attributeName] = await (0, exports.getCroppedText)(current.content, maxTokenSize, tokens, promptConfig.wishList.beforeCursor.strategy, tokenizer, languageParser, (_a = wishList.beforeCursor.extOption.minBlockSize) !== null && _a !== void 0 ? _a : 75, false, token);
                if (token.isCancellationRequested) {
                    return;
                }
                leftTokenSize -= maxTokenSize;
            }
            else {
                leftTokenSize -= currentTokenSize;
                promptList[current.attributeName] = `${current.content}`;
            }
        }
        else if (current.importedFiles && current.importedFiles.length > 0) {
            let currentImportedFile = '';
            const maxImportedFileTokenSize = Math.ceil(current.maxPercent * leftTokenSize);
            let currentImportedFileTokenSize = 0;
            for (const [filename, importedFile] of current.importedFiles) {
                const importedFileMarker = (0, exports.getMarkerByLanguage)(`${promptConfig.wishList.importedFile.extOption.patternPrefix}${filename}:${promptConfig.wishList.importedFile.extOption.patternSuffix || '\n'}${(0, exports.getMarkerForSnippets)(importedFile.join('\n'), context.language)}\n`, context.language);
                const tokens = tokenizer.encode(importedFileMarker);
                if (tokens.length > leftTokenSize || currentImportedFileTokenSize + tokens.length > maxImportedFileTokenSize) {
                    break;
                }
                leftTokenSize -= tokens.length;
                currentImportedFileTokenSize += tokens.length;
                currentImportedFile += importedFileMarker;
            }
            if (currentImportedFile) {
                // 移除末尾多余的 \n
                currentImportedFile = currentImportedFile.slice(0, -'\n'.length);
                promptList[current.attributeName] = `${currentImportedFile}`;
            }
        }
        else if (current.similarSnippets && current.similarSnippets.length > 0) {
            let currentSnippet = '';
            const maxSnippetSize = Math.ceil(current.maxPercent * leftTokenSize);
            let currentSnippetSize = 0;
            for (const [filename, snippets] of current.similarSnippets) {
                const similarSnippetMarker = snippets
                    .map((snippet) => (0, exports.getMarkerByLanguage)(`${promptConfig.wishList.similarFile.extOption.patternPrefix}${filename}:${promptConfig.wishList.similarFile.extOption.patternSuffix || '\n'}${(0, exports.getMarkerForSnippets)(snippet.snippet, context.language)}`, context.language))
                    .join('\n');
                const tokens = tokenizer.encode(similarSnippetMarker);
                if (tokens.length > leftTokenSize || currentSnippetSize + tokens.length > maxSnippetSize) {
                    break;
                }
                leftTokenSize -= tokens.length;
                currentSnippetSize += tokens.length;
                currentSnippet += similarSnippetMarker;
            }
            if (currentSnippet) {
                promptList[current.attributeName] = `${currentSnippet}`;
            }
        }
    }
    return `${(0, exports.addComment)(promptList.languageMarker)}${(0, exports.addComment)(promptList.pathMarker)}${(0, exports.addComment)(promptList.importedFile)}${(0, exports.addComment)(promptList.similarFile)}${(0, exports.addComment)(promptList.beforeCursor)}`.slice(0, -'\n'.length);
};
exports.getBeforePrompt = getBeforePrompt;
const getAfterPrompt = async (context, promptConfig, injector, token) => {
    var _a;
    const { maxPromptTokenSize, tokenizerName, wishList } = promptConfig;
    const tokenizer = (0, tokenizer_1.getTokenizer)(tokenizerName);
    let afterCursor = context.suffix;
    const afterCursorOptions = promptConfig.wishList.afterCursor.extOption;
    const afterCursorTokens = afterCursor.split('\n').map((line) => tokenizer.encode(line));
    const currentAfterCursorTokenSize = afterCursorTokens.reduce((prev, cur) => prev + cur.length, 0);
    const afterCursorMaxTokenSize = Math.ceil(afterCursorOptions.suffixPercent * maxPromptTokenSize);
    if (currentAfterCursorTokenSize > afterCursorMaxTokenSize) {
        const languageParserService = injector.get(service_1.LanguageParserService);
        const languageParser = languageParserService.createParser(context.language);
        afterCursor = await (0, exports.getCroppedText)(afterCursor, afterCursorMaxTokenSize, afterCursorTokens, promptConfig.wishList.afterCursor.strategy, tokenizer, languageParser, (_a = wishList.afterCursor.extOption.minBlockSize) !== null && _a !== void 0 ? _a : 25, true, token);
    }
    return afterCursor;
};
exports.getAfterPrompt = getAfterPrompt;
//# sourceMappingURL=prompt.js.map