"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.LanguageParser = exports.DEFAULT_MIN_BLOCK_COUNT = void 0;
const tslib_1 = require("tslib");
const web_tree_sitter_1 = tslib_1.__importDefault(require("web-tree-sitter"));
const di_1 = require("@opensumi/di");
const ide_utils_1 = require("@opensumi/ide-utils");
const common_1 = require("./tree-sitter/common");
const language_facts_1 = require("./tree-sitter/language-facts");
const wasm_manager_1 = require("./tree-sitter/wasm-manager");
exports.DEFAULT_MIN_BLOCK_COUNT = 20;
let LanguageParser = class LanguageParser {
    constructor(language) {
        this.language = language;
        this.parserLoaded = new ide_utils_1.Deferred();
        this.lruCache = new ide_utils_1.LRUCache(500);
        this.initializeParser();
    }
    ready() {
        return this.parserLoaded.promise;
    }
    getParser() {
        return this.parser;
    }
    async initializeParser() {
        this.parser = await this.wasmModuleManager.initParser();
        // Load grammar
        const grammar = await this.wasmModuleManager.loadLanguage(this.language);
        const languageParser = await web_tree_sitter_1.default.Language.load(new Uint8Array(grammar));
        // Set language
        this.parser.setLanguage(languageParser);
        this.parserLoaded.resolve();
    }
    /**
     * 从给定的位置开始，向上遍历 AST，找到最近的代码块
     * @param node 节点
     * @param position 位置
     * @returns 代码块
     */
    findContainingCodeBlockWithPosition(node, position) {
        if (node.startIndex <= position && node.endIndex >= position) {
            const isBlockIdentifier = this.languageFacts.isCodeBlock(this.language, node.type);
            if (isBlockIdentifier) {
                return node;
            }
        }
        if (node.parent) {
            return this.findContainingCodeBlockWithPosition(node.parent, position);
        }
        return null;
    }
    findFunctionCodeBlock(node, position) {
        if (node.startIndex <= position && node.endIndex >= position) {
            const valid = this.languageFacts.isFunctionCodeBlock(this.language, node.type);
            if (valid) {
                return node;
            }
        }
        if (node.parent) {
            return this.findFunctionCodeBlock(node.parent, position);
        }
        return null;
    }
    async parseAST(model) {
        const key = `${model.id}@${model.getVersionId()}`;
        const cachedNode = this.lruCache.get(key);
        if (cachedNode) {
            return cachedNode;
        }
        await this.parserLoaded.promise;
        const sourceCode = model.getValue();
        const tree = this.parser.parse(sourceCode);
        if (tree) {
            const rootNode = tree.rootNode;
            this.lruCache.set(key, rootNode);
            return rootNode;
        }
    }
    async getSyntaxNodeAsPosition(model, cursor) {
        const rootNode = await this.parseAST(model);
        if (rootNode) {
            const cursorNode = rootNode.namedDescendantForIndex(cursor);
            return cursorNode;
        }
        return null;
    }
    /**
     * 从给定的位置开始，找到最近的没有语法错误的代码块
     */
    async findCodeBlockWithSyntaxError(sourceCode, range) {
        await this.parserLoaded.promise;
        const tree = this.parser.parse(sourceCode);
        if (tree) {
            const rootNode = tree.rootNode;
            const startPosition = {
                row: range.startLineNumber - 1,
                column: range.startColumn,
            };
            const selectedNode = rootNode.namedDescendantForPosition(startPosition);
            let parentNode = selectedNode.parent;
            if (!parentNode) {
                return {
                    range: (0, common_1.toMonacoRange)(selectedNode),
                    type: selectedNode.type,
                    infoCategory: 'other',
                };
            }
            // 检查父节点是否有语法错误，并向上遍历
            while (parentNode) {
                // 检查当前节点是否有错误
                const hasError = parentNode.hasError;
                // 如果有错误，停止遍历，当前节点的父节点即为所求的代码块
                if (hasError) {
                    break;
                }
                // 向上移动到父节点
                parentNode = parentNode.parent;
            }
            if (parentNode) {
                return {
                    range: (0, common_1.toMonacoRange)(parentNode),
                    type: parentNode.type,
                    infoCategory: 'other',
                };
            }
            return {
                range,
                type: selectedNode.type,
                infoCategory: 'other',
            };
        }
        return null;
    }
    async provideCodeBlockInfo(model, position) {
        const cursor = model.getOffsetAt(position);
        const cursorNode = await this.getSyntaxNodeAsPosition(model, cursor);
        if (!cursorNode) {
            return null;
        }
        const functionNode = this.findFunctionCodeBlock(cursorNode, cursor);
        if (functionNode) {
            return this.languageFacts.provideFunctionInfo(this.language, functionNode);
        }
        const selectedNode = this.findContainingCodeBlockWithPosition(cursorNode, cursor);
        if (selectedNode) {
            return {
                infoCategory: 'other',
                range: {
                    startLineNumber: selectedNode.startPosition.row + 1,
                    startColumn: 0,
                    endLineNumber: selectedNode.endPosition.row + 1,
                    endColumn: Infinity,
                },
                type: selectedNode.type,
            };
        }
        return null;
    }
    async provideCodeBlockInfoInRange(model, range) {
        const rootNode = await this.parseAST(model);
        if (rootNode) {
            const startPosition = {
                row: range.startLineNumber - 1,
                column: range.startColumn - 1,
            };
            const endPosition = {
                row: range.endLineNumber,
                column: range.endColumn,
            };
            const types = this.languageFacts.getCodeBlockTypes(this.language);
            if (!types || types.size === 0) {
                return null;
            }
            const nodes = rootNode.descendantsOfType(Array.from(types), startPosition, endPosition);
            if (nodes && nodes.length > 0) {
                const firstNode = nodes[0];
                const range = (0, common_1.toMonacoRange)(firstNode);
                return {
                    infoCategory: 'other',
                    range,
                    type: firstNode.type,
                };
            }
        }
        return null;
    }
    async trimSuffixSyntaxErrors(code, minBlockCount = exports.DEFAULT_MIN_BLOCK_COUNT) {
        var _a;
        await this.parserLoaded.promise;
        const tree = (_a = this.parser) === null || _a === void 0 ? void 0 : _a.parse(code);
        const rootNode = tree === null || tree === void 0 ? void 0 : tree.rootNode;
        if (!rootNode) {
            return code;
        }
        const { namedChildren } = rootNode;
        if (!rootNode.hasError || namedChildren.length <= minBlockCount) {
            return this.trimToLastCompleteBlock(rootNode, true);
        }
        let lastErrorNode = null;
        let targetNode = null;
        for (let i = namedChildren.length - 1; i >= minBlockCount; i--) {
            const child = namedChildren[i];
            if (child.hasError && i !== 0) {
                lastErrorNode = child;
            }
            else if (lastErrorNode) {
                targetNode = child;
                break;
            }
        }
        if (!targetNode) {
            targetNode = namedChildren[minBlockCount - 1];
        }
        const lines = code.split('\n');
        const remainingLines = lines.slice(0, targetNode.endPosition.row + 1);
        const lastLine = remainingLines[remainingLines.length - 1].slice(0, targetNode.endPosition.column);
        const correctCode = `${remainingLines.length > 1 ? `${remainingLines.slice(0, -1).join('\n')}\n` : ''}${lastLine}`;
        return correctCode;
    }
    async trimPrefixSyntaxErrors(code, minBlockCount = exports.DEFAULT_MIN_BLOCK_COUNT) {
        var _a;
        await this.parserLoaded.promise;
        const tree = (_a = this.parser) === null || _a === void 0 ? void 0 : _a.parse(code);
        const rootNode = tree === null || tree === void 0 ? void 0 : tree.rootNode;
        if (!rootNode) {
            return code;
        }
        const { namedChildren } = rootNode;
        if (!rootNode.hasError || namedChildren.length <= minBlockCount) {
            return this.trimToLastCompleteBlock(rootNode);
        }
        let firstErrorNode = null;
        let targetNode = null;
        for (let i = 0; i < namedChildren.length - minBlockCount; i++) {
            const child = namedChildren[i];
            if (child.hasError && i !== namedChildren.length - 1) {
                firstErrorNode = child;
            }
            else if (firstErrorNode) {
                targetNode = child;
                break;
            }
        }
        if (!targetNode) {
            targetNode = namedChildren[namedChildren.length - minBlockCount];
        }
        const lines = code.split('\n');
        const remainingLines = lines.slice(targetNode.startPosition.row);
        const firstLine = remainingLines[0].slice(targetNode.startPosition.column);
        const correctCode = `${firstLine}${remainingLines.length > 1 ? `\n${remainingLines.slice(1).join('\n')}` : ''}`;
        return correctCode;
    }
    trimToLastCompleteBlock(rootNode, reverse = false) {
        const { namedChildren } = rootNode;
        if (reverse) {
            const lastNamedChild = namedChildren[namedChildren.length - 2];
            if (lastNamedChild) {
                const lines = rootNode.text.split('\n');
                const remainingLines = lines.slice(0, lastNamedChild.endPosition.row + 1);
                const lastLine = remainingLines[remainingLines.length - 1].slice(0, lastNamedChild.endPosition.column);
                const correctCode = `${remainingLines.length > 1 ? `${remainingLines.slice(0, -1).join('\n')}\n` : ''}${lastLine}`;
                return correctCode;
            }
        }
        else {
            const firstNamedChild = namedChildren[1];
            if (firstNamedChild) {
                const lines = rootNode.text.split('\n');
                const remainingLines = lines.slice(firstNamedChild.startPosition.row);
                const firstLine = remainingLines[0].slice(firstNamedChild.startPosition.column);
                const correctCode = `${firstLine}${remainingLines.length > 1 ? `\n${remainingLines.slice(1).join('\n')}` : ''}`;
                return correctCode;
            }
        }
        return rootNode.text;
    }
    async extractImportPaths(code) {
        var _a, _b;
        const paths = [];
        if (this.language === 'typescript') {
            await this.parserLoaded.promise;
            const tree = (_a = this.parser) === null || _a === void 0 ? void 0 : _a.parse(code);
            const rootNode = tree === null || tree === void 0 ? void 0 : tree.rootNode;
            if (rootNode) {
                for (let i = 0; i < (rootNode === null || rootNode === void 0 ? void 0 : rootNode.childCount); i++) {
                    const sourceChild = rootNode.child(i);
                    if ((sourceChild === null || sourceChild === void 0 ? void 0 : sourceChild.type) === 'import_statement') {
                        let importPath = (_b = sourceChild.child(3)) === null || _b === void 0 ? void 0 : _b.text;
                        if ((importPath === null || importPath === void 0 ? void 0 : importPath.includes("'")) || (importPath === null || importPath === void 0 ? void 0 : importPath.includes('"'))) {
                            importPath = importPath.slice(1, -1);
                        }
                        if (importPath) {
                            paths.push(importPath);
                        }
                    }
                }
            }
        }
        return paths;
    }
    async extractInterfaceOrTypeCode(code) {
        var _a;
        const snippets = [];
        if (this.language === 'typescript') {
            await this.parserLoaded.promise;
            const tree = (_a = this.parser) === null || _a === void 0 ? void 0 : _a.parse(code);
            const rootNode = tree === null || tree === void 0 ? void 0 : tree.rootNode;
            if (rootNode) {
                for (let i = 0; i < (rootNode === null || rootNode === void 0 ? void 0 : rootNode.childCount); i++) {
                    const sourceChild = rootNode.child(i);
                    if ((sourceChild === null || sourceChild === void 0 ? void 0 : sourceChild.type) === 'export_statement') {
                        const exportChild = sourceChild.child(1);
                        if ((exportChild === null || exportChild === void 0 ? void 0 : exportChild.type) === 'interface_declaration' || (exportChild === null || exportChild === void 0 ? void 0 : exportChild.type) === 'type_alias_declaration') {
                            snippets.push(exportChild.text);
                        }
                    }
                    else if ((sourceChild === null || sourceChild === void 0 ? void 0 : sourceChild.type) === 'interface_declaration' || (sourceChild === null || sourceChild === void 0 ? void 0 : sourceChild.type) === 'type_alias_declaration') {
                        snippets.push(sourceChild.text);
                    }
                }
            }
        }
        return snippets;
    }
    dispose() {
        this.parser.delete();
        this.lruCache.clear();
    }
};
tslib_1.__decorate([
    (0, di_1.Autowired)(wasm_manager_1.WasmModuleManager),
    tslib_1.__metadata("design:type", wasm_manager_1.WasmModuleManager)
], LanguageParser.prototype, "wasmModuleManager", void 0);
tslib_1.__decorate([
    (0, di_1.Autowired)(language_facts_1.TreeSitterLanguageFacts),
    tslib_1.__metadata("design:type", language_facts_1.TreeSitterLanguageFacts)
], LanguageParser.prototype, "languageFacts", void 0);
LanguageParser = tslib_1.__decorate([
    (0, di_1.Injectable)({ multiple: true }),
    tslib_1.__metadata("design:paramtypes", [String])
], LanguageParser);
exports.LanguageParser = LanguageParser;
//# sourceMappingURL=parser.js.map